/*==========================================================================
 *
 *  Copyright (C) 2001 Crocus. All Rights Reserved.
 *
 *  File:       dsutil.cpp
 *  Content:    Routines pour lire les fichier son.
 *	Base:		Dsutil.cpp fourni pour Direct Sound
 *
 ***************************************************************************/

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#include <mmsystem.h>
#include <dsound.h>

typedef struct{
	void *MemAlloc;		// Ajout
	BYTE *pbWaveData;
	DWORD cbWaveSize;
	int iAlloc;
	int iCurrent;
	IDirectSoundBuffer* Buffers[1];
}SNDOBJ, *HSNDOBJ;
#define _HSNDOBJ_DEFINED

#include "dsutil.h"

///////////////////////////////////////////////////////////////////////////////
//
// DSLoadSoundBuffer (Modifi)
//
///////////////////////////////////////////////////////////////////////////////

IDirectSoundBuffer *DSLoadSoundBuffer(IDirectSound *pDS, void *MemAlloc){
    IDirectSoundBuffer *pDSB = NULL;
    DSBUFFERDESC dsBD = {0};
    BYTE *pbWaveData;
	
    if(DSParseWaveFile(MemAlloc, &dsBD.lpwfxFormat, &pbWaveData, &dsBD.dwBufferBytes)){
        dsBD.dwSize = sizeof(dsBD);
        dsBD.dwFlags = DSBCAPS_STATIC | DSBCAPS_CTRLPAN | DSBCAPS_CTRLVOLUME | DSBCAPS_CTRLFREQUENCY;
		
        if(SUCCEEDED(IDirectSound_CreateSoundBuffer(pDS, &dsBD, &pDSB, NULL))){
            if (!DSFillSoundBuffer(pDSB, pbWaveData, dsBD.dwBufferBytes)){
                IDirectSoundBuffer_Release(pDSB);
                pDSB = NULL;
            }
        }else pDSB = NULL;
    }

    return pDSB;
}

///////////////////////////////////////////////////////////////////////////////
//
// DSReloadSoundBuffer (Modifi)
//
///////////////////////////////////////////////////////////////////////////////

BOOL DSReloadSoundBuffer(IDirectSoundBuffer *pDSB, void *MemAlloc){
	BOOL result=FALSE;
	BYTE *pbWaveData;
	DWORD cbWaveSize;

	if (DSParseWaveFile(MemAlloc, NULL, &pbWaveData, &cbWaveSize))
    	if (SUCCEEDED(IDirectSoundBuffer_Restore(pDSB)) && DSFillSoundBuffer(pDSB, pbWaveData, cbWaveSize))
			result = TRUE;

	return result;
}

///////////////////////////////////////////////////////////////////////////////
//
// DSGetWaveFile (Nouveau) DSGetWaveResource (supprimer)
//
///////////////////////////////////////////////////////////////////////////////

void *DSGetWaveFile(LPCTSTR lpName,WAVEFORMATEX **ppWaveHeader, BYTE **ppbWaveData, DWORD *pcbWaveSize){
    void *pvRes;
	int dwSize, nbLus;

	HANDLE hFich = CreateFile(lpName, GENERIC_READ, 0, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
	if(hFich==INVALID_HANDLE_VALUE) return NULL;
	dwSize = SetFilePointer(hFich, 0, NULL, FILE_END);
	SetFilePointer(hFich,0, NULL, FILE_BEGIN);
	pvRes = malloc(dwSize);
	ReadFile(hFich, pvRes, dwSize, &nbLus, 0);
	CloseHandle(hFich);

	DSParseWaveFile(pvRes, ppWaveHeader, ppbWaveData, pcbWaveSize);
	return pvRes;
}

///////////////////////////////////////////////////////////////////////////////
// SndObj fns (Modifi)
///////////////////////////////////////////////////////////////////////////////

SNDOBJ *SndObjCreate(IDirectSound *pDS, LPCTSTR lpName, int iConcurrent){
    SNDOBJ *pSO = NULL;
    LPWAVEFORMATEX pWaveHeader;
    BYTE *pbData;
    UINT cbData;
	void *MemAlloc;
	int i;

	MemAlloc = DSGetWaveFile(lpName, &pWaveHeader, &pbData, &cbData);
    if(MemAlloc != NULL){
        if (iConcurrent < 1) iConcurrent = 1;
		pSO=(SNDOBJ*)LocalAlloc(LPTR,sizeof(SNDOBJ)+(iConcurrent-1)*sizeof(IDirectSoundBuffer *));
        if (pSO!=NULL){
			pSO->MemAlloc = MemAlloc;
            pSO->iAlloc = iConcurrent;
            pSO->pbWaveData = pbData;
            pSO->cbWaveSize = cbData;
            pSO->Buffers[0] = DSLoadSoundBuffer(pDS, MemAlloc);
            for (i=1; i<pSO->iAlloc; i++)
                if(FAILED(IDirectSound_DuplicateSoundBuffer(pDS,pSO->Buffers[0], &pSO->Buffers[i]))){
                    pSO->Buffers[i] = DSLoadSoundBuffer(pDS, MemAlloc);
                    if(!pSO->Buffers[i]){
                        SndObjDestroy(pSO);
                        pSO = NULL;
                        break;
                    }
                }
        }
    }
    return pSO;
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

void SndObjDestroy(SNDOBJ *pSO){
	int i;
	if(pSO){
		SndObjStop(pSO);
		for(i=0;i<pSO->iAlloc;i++) if(pSO->Buffers[i]) IDirectSoundBuffer_Release(pSO->Buffers[i]);
		free(pSO->MemAlloc);
        LocalFree((HANDLE)pSO);
    }
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

IDirectSoundBuffer *SndObjGetFreeBuffer(SNDOBJ *pSO){
    IDirectSoundBuffer *pDSB;

    if(pSO == NULL) return NULL;

    if(pDSB = pSO->Buffers[pSO->iCurrent]){
        HRESULT hres;
        DWORD dwStatus;

        hres = IDirectSoundBuffer_GetStatus(pDSB, &dwStatus);

        if(FAILED(hres)) dwStatus = 0;

        if((dwStatus & DSBSTATUS_PLAYING) == DSBSTATUS_PLAYING){
            if(pSO->iAlloc > 1){
                if(++pSO->iCurrent >= pSO->iAlloc) pSO->iCurrent = 0;

                pDSB = pSO->Buffers[pSO->iCurrent];
                hres = IDirectSoundBuffer_GetStatus(pDSB, &dwStatus);

                if(SUCCEEDED(hres) && (dwStatus & DSBSTATUS_PLAYING) == DSBSTATUS_PLAYING){
                    IDirectSoundBuffer_Stop(pDSB);
                    IDirectSoundBuffer_SetCurrentPosition(pDSB, 0);
                }
            }else pDSB = NULL;
        }
        if(pDSB && (dwStatus & DSBSTATUS_BUFFERLOST))
            if(FAILED(IDirectSoundBuffer_Restore(pDSB)) || !DSFillSoundBuffer(pDSB, pSO->pbWaveData, pSO->cbWaveSize))
                pDSB = NULL;
    }
    return pDSB;
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

BOOL SndObjPlay(SNDOBJ *pSO, DWORD dwPlayFlags){
    BOOL result = FALSE;

    if(pSO == NULL) return FALSE;

	#ifdef PAS_DE_SON
		return TRUE;
	#endif

    if((!(dwPlayFlags & DSBPLAY_LOOPING) || (pSO->iAlloc == 1))){
        IDirectSoundBuffer *pDSB = SndObjGetFreeBuffer(pSO);
        if (pDSB != NULL) result = SUCCEEDED(IDirectSoundBuffer_Play(pDSB, 0, 0, dwPlayFlags));
    }

    return result;
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

BOOL SndObjStop(SNDOBJ *pSO){
    int i;

    if (pSO == NULL) return FALSE;

    for (i=0; i<pSO->iAlloc; i++){
        IDirectSoundBuffer_Stop(pSO->Buffers[i]);
        IDirectSoundBuffer_SetCurrentPosition(pSO->Buffers[i], 0);
    }

    return TRUE;
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

BOOL DSFillSoundBuffer(IDirectSoundBuffer *pDSB, BYTE *pbWaveData, DWORD cbWaveSize){
    if (pDSB && pbWaveData && cbWaveSize){
        LPVOID pMem1, pMem2;
        DWORD dwSize1, dwSize2;
        if (SUCCEEDED(IDirectSoundBuffer_Lock(pDSB, 0, cbWaveSize, &pMem1, &dwSize1, &pMem2, &dwSize2, 0))){
            CopyMemory(pMem1, pbWaveData, dwSize1);
            if ( 0 != dwSize2 ) CopyMemory(pMem2, pbWaveData+dwSize1, dwSize2);
            IDirectSoundBuffer_Unlock(pDSB, pMem1, dwSize1, pMem2, dwSize2);
            return TRUE;
        }
    }
    return FALSE;
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

BOOL DSParseWaveFile(void *pvRes, WAVEFORMATEX **ppWaveHeader, BYTE **ppbWaveData,DWORD *pcbWaveSize){
    DWORD *pdw;
    DWORD *pdwEnd;
    DWORD dwRiff;
    DWORD dwType;
    DWORD dwLength;

    if (ppWaveHeader)
        *ppWaveHeader = NULL;

    if (ppbWaveData)
        *ppbWaveData = NULL;

    if (pcbWaveSize)
        *pcbWaveSize = 0;

    pdw = (DWORD *)pvRes;
    dwRiff = *pdw++;
    dwLength = *pdw++;
    dwType = *pdw++;

    if(dwRiff != mmioFOURCC('R','I','F','F')) goto exit;
    if(dwType != mmioFOURCC('W','A','V','E')) goto exit;
    pdwEnd = (DWORD *)((BYTE *)pdw + dwLength-4);
    while (pdw < pdwEnd){
        dwType = *pdw++;
        dwLength = *pdw++;

        switch (dwType){
        case mmioFOURCC('f', 'm', 't', ' '):
            if (ppWaveHeader && !*ppWaveHeader){
                if (dwLength < sizeof(WAVEFORMAT)) goto exit;
                *ppWaveHeader = (WAVEFORMATEX *)pdw;
                if ((!ppbWaveData || *ppbWaveData) && (!pcbWaveSize || *pcbWaveSize)) return TRUE;
			}
            break;

        case mmioFOURCC('d', 'a', 't', 'a'):
            if ((ppbWaveData && !*ppbWaveData) || (pcbWaveSize && !*pcbWaveSize)){
                if (ppbWaveData) *ppbWaveData = (LPBYTE)pdw;
                if (pcbWaveSize) *pcbWaveSize = dwLength;
                if (!ppWaveHeader || *ppWaveHeader) return TRUE;
            }
            break;
        }
        pdw = (DWORD *)((BYTE *)pdw + ((dwLength+1)&~1));
    }

exit:
    return FALSE;
}
